//! # Portfolio
//!
//! $$
//! \sigma_p^2 = \mathbf{w}^\top \Sigma \mathbf{w}
//! $$
//!
//! Portfolio optimization and momentum portfolio construction.

pub mod data;
pub mod engine;
pub mod momentum;
pub mod optimizers;
pub mod types;
